@@ -29,6 +29,7 @@
 #include <unistd.h>
 
 #include <glib/gstdio.h>
+#include <glib/gi18n.h>
 
 #include "seahorse-util.h"
 #include "seahorse-libdialogs.h"
@@ -78,7 +79,7 @@
  * @email: Optional user ID email
  * @comment: Optional user ID comment
  * @passphrase: Passphrase for key
- * @type: Key type. Supported types are #DSA_ELGAMAL, #DSA, and #RSA_SIGN
+ * @type: Key type. Supported types are #DSA_ELGAMAL, #DSA, #RSA_SIGN, and #RSA_RSA
  * @length: Length of key, must be within the range of @type specified by #SeahorseKeyLength
  * @expires: Expiration date of key
  * @err: Catches errors in the params
@@ -112,7 +113,7 @@ seahorse_gpgme_key_op_generate (SeahorseGpgmeSource *psrc, const gchar *name,
         if (length < DSA_MIN || length > DSA_MAX)
             *err = GPG_E (GPG_ERR_INV_VALUE);
         break;
-    case RSA_SIGN:
+    case RSA_RSA: RSA_SIGN:
         if (length < RSA_MIN || length > LENGTH_MAX)
             *err = GPG_E (GPG_ERR_INV_VALUE);
         break;
@@ -134,10 +135,10 @@ seahorse_gpgme_key_op_generate (SeahorseGpgmeSource *psrc, const gchar *name,
     if (comment != NULL && strlen (comment) > 0)
         common = g_strdup_printf ("Name-Comment: %s\n%s", comment, common);
 
-    if (type == RSA_SIGN)
-        key_type = "Key-Type: RSA\nKey-Usage: sign";
-    else
+    if (type == DSA || type == DSA_ELGAMAL)
         key_type = "Key-Type: DSA\nKey-Usage: sign";
+    else
+        key_type = "Key-Type: RSA\nKey-Usage: sign";
 
     start = g_strdup_printf ("<GnupgKeyParms format=\"internal\">\n%s\nKey-Length: ", key_type);
 
@@ -145,6 +146,9 @@ seahorse_gpgme_key_op_generate (SeahorseGpgmeSource *psrc, const gchar *name,
     if (type == DSA_ELGAMAL)
         parms = g_strdup_printf ("%s%d\nSubkey-Type: ELG-E\nSubkey-Length: %d\nSubkey-Usage: encrypt\n%s",
                                  start, (length < DSA_MAX) ? length : DSA_MAX, length, common);
+    else if (type == RSA_RSA)
+        parms = g_strdup_printf ("%s%d\nSubkey-Type: RSA\nSubkey-Length: %d\nSubkey-Usage: encrypt\n%s",
+                                 start, length, length, common);
     else
         parms = g_strdup_printf ("%s%d\n%s", start, length, common);
 
@@ -291,6 +295,10 @@ edit_gpgme_key (gpgme_ctx_t ctx, gpgme_key_t key, SeahorseEditParm *parms)
     
 	/* do edit callback, release data */
 	gerr = gpgme_op_edit (ctx, key, seahorse_gpgme_key_op_edit, parms, out);
+
+	if (gpgme_err_code (gerr) == GPG_ERR_BAD_PASSPHRASE) {
+		seahorse_util_show_error(NULL, _("Wrong password"), _("This was the third time you entered a wrong password. Please try again."));
+	}
 	
 	seahorse_gpgme_data_release (out);
     
@@ -640,7 +648,7 @@ edit_pass_transit (guint current_state, gpgme_status_code_t status,
         
         /* If all tries for passphrase were wrong, we get here */
         else if (status == GPGME_STATUS_GET_LINE && g_str_equal (args, PROMPT)) {
-            *err = GPG_E (GPG_ERR_CANCELED);
+            *err = GPG_E (GPG_ERR_BAD_PASSPHRASE);
             next_state = PASS_ERROR;
         
         /* No idea how we got here ... */
@@ -1389,7 +1397,7 @@ typedef enum {
 } AddKeyState;
 
 typedef struct {
-	SeahorseKeyEncType  type;
+	guint               type;
 	guint               length;
 	time_t              expires;
 } SubkeyParm;
@@ -1495,19 +1503,31 @@ seahorse_gpgme_key_op_add_subkey (SeahorseGpgmeKey *pkey, const SeahorseKeyEncTy
 {
 	SeahorseEditParm *parms;
 	SubkeyParm *key_parm;
+	guint real_type;
+	SeahorseKeyTypeTable table;
+	gpgme_error_t gerr;
 	
     g_return_val_if_fail (SEAHORSE_IS_GPGME_KEY (pkey), GPG_E (GPG_ERR_WRONG_KEY_USAGE));    
     g_return_val_if_fail (seahorse_object_get_usage (SEAHORSE_OBJECT (pkey)) == SEAHORSE_USAGE_PRIVATE_KEY, GPG_E (GPG_ERR_WRONG_KEY_USAGE));
 	
+	gerr = seahorse_gpgme_get_keytype_table (&table);
+	g_return_val_if_fail (GPG_IS_OK (gerr), gerr);
+	
 	/* Check length range & type */
 	switch (type) {
 		case DSA:
+			real_type = table->dsa_sign;
 			g_return_val_if_fail (length >= DSA_MIN && length <= DSA_MAX, GPG_E (GPG_ERR_INV_VALUE));
 			break;
 		case ELGAMAL:
+			real_type = table->elgamal_enc;
 			g_return_val_if_fail (length >= ELGAMAL_MIN && length <= LENGTH_MAX, GPG_E (GPG_ERR_INV_VALUE));
 			break;
 		case RSA_SIGN: case RSA_ENCRYPT:
+			if (type == RSA_SIGN)
+				real_type = table->rsa_sign;
+			else
+				real_type = table->rsa_enc;
 			g_return_val_if_fail (length >= RSA_MIN && length <= LENGTH_MAX, GPG_E (GPG_ERR_INV_VALUE));
 			break;
 		default:
@@ -1516,7 +1536,7 @@ seahorse_gpgme_key_op_add_subkey (SeahorseGpgmeKey *pkey, const SeahorseKeyEncTy
 	}
 	
 	key_parm = g_new (SubkeyParm, 1);
-	key_parm->type = type;
+	key_parm->type = real_type;
 	key_parm->length = length;
 	key_parm->expires = expires;
 	
